/*
 * Copyright (C) 2004 Nathan Lutchansky <lutchann@litech.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

#include <event.h>
#include <log.h>
#include <frame.h>
#include <rtp.h>

void random_bytes( unsigned char *dest, int len );
int interleave_send( void *conn, int chan, unsigned char *buf, int len );

static void init_rtp_endpoint( struct rtp_endpoint *ep )
{
	ep->next = ep->rtp_stream->ep_list;
	ep->prev = NULL;
	if( ep->next ) ep->next->prev = ep;
	ep->rtp_stream->ep_list = ep;
	ep->payload = 96;
	random_bytes( (unsigned char *)&ep->ssrc, 4 );
	random_bytes( (unsigned char *)&ep->timestamp, 4 );
	random_bytes( (unsigned char *)&ep->seqnum, 2 );
	ep->packet_count = 0;
	ep->octet_count = 0;
	ep->running = 0;
}

void del_rtp_endpoint( struct rtp_endpoint *ep )
{
	if( ep->session && ep->session->endpoint == ep )
		ep->session->endpoint = NULL;

	if( ep->running )
	{
		ep->running = 0;
		ep->rtp_stream->del_endpoint( ep->rtp_stream, ep );
	}
	if( ep->next ) ep->next->prev = ep->prev;
	if( ep->prev ) ep->prev->next = ep->next;
	else ep->rtp_stream->ep_list = ep->next;

	switch( ep->trans_type )
	{
	case RTP_TRANS_UDP:
		remove_event( ep->trans.udp.rtp_event );
		close( ep->trans.udp.rtp_fd );
		remove_event( ep->trans.udp.rtcp_event );
		close( ep->trans.udp.rtcp_fd );
		break;
	}

	free( ep );
}

void update_rtp_timestamp( struct rtp_stream *rtp, int time_increment )
{
	struct rtp_endpoint *ep;

	for( ep = rtp->ep_list; ep; ep = ep->next )
	{
		ep->timestamp += time_increment;
		ep->timestamp &= 0xFFFFFFFF;
	}
}

static void udp_rtp_read( struct event *e, void *d )
{
	struct rtp_endpoint *ep = (struct rtp_endpoint *)d;
	unsigned char buf[16384];
	int ret;

	ret = read( ep->trans.udp.rtp_fd, buf, sizeof( buf ) );
	if( ret > 0 ) return;
	if( ret < 0 )
		spook_log( SL_VERBOSE, "error on UDP RTP socket: %s",
			strerror( errno ) );
	else spook_log( SL_VERBOSE, "UDP RTP socket closed" );
	del_rtp_endpoint( ep );
}

static void udp_rtcp_read( struct event *e, void *d )
{
	struct rtp_endpoint *ep = (struct rtp_endpoint *)d;
	unsigned char buf[16384];
	int ret;

	ret = read( ep->trans.udp.rtcp_fd, buf, sizeof( buf ) );
	if( ret > 0 ) return;
	if( ret < 0 )
		spook_log( SL_VERBOSE, "error on UDP RTCP socket: %s",
			strerror( errno ) );
	else spook_log( SL_VERBOSE, "UDP RTCP socket closed" );
	del_rtp_endpoint( ep );
}

static void rtcp_send( struct event *e, void *d )
{
	struct rtp_endpoint *ep = (struct rtp_endpoint *)d;
	struct timeval now;
	unsigned char buf[16384];
	unsigned int ntp_sec, ntp_usec;

	gettimeofday( &now, NULL );
	ntp_sec = now.tv_sec + 0x83AA7E80;
	ntp_usec = (double)( (double)now.tv_usec * (double)0x4000000 ) / 15625.0;
	
	//spook_log( SL_DEBUG, "ssrc=%u, ntp_sec=%u, ntp_usec=%u last_timestamp=%u packet_count=%d octet_count=%d",
	//		x->ssrc, ntp_sec, ntp_usec, x->last_timestamp, x->packet_count, x->octet_count );

	buf[0] = 2 << 6; // version
	buf[1] = 200; // packet type is Sender Report
	PUT_16( buf + 2, 6 ); // length in words minus one
	PUT_32( buf + 4, ep->ssrc );
	PUT_32( buf + 8, ntp_sec );
	PUT_32( buf + 12, ntp_usec );
	PUT_32( buf + 16, ep->timestamp );
	PUT_32( buf + 20, ep->packet_count );
	PUT_32( buf + 24, ep->octet_count );
	buf[28] = ( 2 << 6 ) | 1; // version; source count = 1
	buf[29] = 202; // packet type is Source Description
	PUT_16( buf + 30, 4 ); // length in words minus one
	PUT_32( buf + 32, ep->ssrc );
	buf[36] = 0x01; // field type is CNAME
	buf[37] = 14; // text length
	memcpy( buf + 38, "Unnamed stream", 14 );
	switch( ep->trans_type )
	{
	case RTP_TRANS_UDP:
		write( ep->trans.udp.rtcp_fd, buf, 52 );
		break;
	case RTP_TRANS_INTER:
		interleave_send( ep->trans.inter.conn,
				ep->trans.inter.chan + 1, buf, 52 );
		break;
	}
}

struct rtp_endpoint *new_rtp_udp_endpoint( struct rtp_stream *rtp,
		struct in_addr dest_ip, int dest_port, int *our_port )
{
	struct rtp_endpoint *ep;
	struct sockaddr_in rtpaddr, rtcpaddr;
	int port, tries, rtpfd = -1, rtcpfd = -1;

	rtpaddr.sin_family = rtcpaddr.sin_family = AF_INET;
	rtpaddr.sin_addr.s_addr = rtcpaddr.sin_addr.s_addr = 0;

	port = ( 50000 + random() % 10000 ) & ~1;
	for( tries = 0; tries < 10000 / 2; ++tries )
	{
		if( port >= 60000 ) port = 50000;
		rtpaddr.sin_port = htons( port );
		rtcpaddr.sin_port = htons( port + 1 );
		if( rtpfd < 0 &&
			( rtpfd = socket( PF_INET, SOCK_DGRAM, 0 ) ) < 0 )
		{
			spook_log( SL_WARN, "unable to create UDP RTP socket: %s",
					strerror( errno ) );
			return NULL;
		}
		if( rtcpfd < 0 &&
			( rtcpfd = socket( PF_INET, SOCK_DGRAM, 0 ) ) < 0 )
		{
			spook_log( SL_WARN, "unable to create UDP RTCP socket: %s",
					strerror( errno ) );
			close( rtpfd );
			return NULL;
		}
		if( bind( rtpfd, (struct sockaddr *)&rtpaddr,
					sizeof( rtpaddr ) ) < 0 )
		{
			if( errno == EADDRINUSE )
			{
				port += 2;
				continue;
			}
			spook_log( SL_WARN, "strange error when binding RTP socket: %s",
					strerror( errno ) );
			close( rtpfd );
			close( rtcpfd );
			return NULL;
		}
		if( bind( rtcpfd, (struct sockaddr *)&rtcpaddr,
					sizeof( rtcpaddr ) ) < 0 )
		{
			if( errno == EADDRINUSE )
			{
				close( rtpfd );
				rtpfd = -1;
				port += 2;
				continue;
			}
			spook_log( SL_WARN, "strange error when binding RTCP socket: %s",
					strerror( errno ) );
			close( rtpfd );
			close( rtcpfd );
			return NULL;
		}
		break;
	}
	if( tries == 10000 / 2 )
	{
		spook_log( SL_WARN, "ran out of UDP RTP ports!" );
		return NULL;
	}
	rtpaddr.sin_family = rtcpaddr.sin_family = AF_INET;
	rtpaddr.sin_addr = rtcpaddr.sin_addr = dest_ip;
	rtpaddr.sin_port = htons( dest_port );
	rtcpaddr.sin_port = htons( dest_port + 1 );
	if( connect( rtpfd, (struct sockaddr *)&rtpaddr,
				sizeof( rtpaddr ) ) < 0 )
	{
		spook_log( SL_WARN, "strange error when connecting RTP socket: %s",
				strerror( errno ) );
		close( rtpfd );
		close( rtcpfd );
		return NULL;
	}
	if( connect( rtcpfd, (struct sockaddr *)&rtcpaddr,
				sizeof( rtcpaddr ) ) < 0 )
	{
		spook_log( SL_WARN, "strange error when connecting RTCP socket: %s",
				strerror( errno ) );
		close( rtpfd );
		close( rtcpfd );
		return NULL;
	}

	ep = (struct rtp_endpoint *)malloc( sizeof( struct rtp_endpoint ) );
	ep->rtp_stream = rtp;
	ep->trans_type = RTP_TRANS_UDP;
	ep->trans.udp.rtp_fd = rtpfd;
	ep->trans.udp.rtcp_fd = rtcpfd;
	ep->trans.udp.rtp_event = add_fd_event( rtpfd, 0, 0, udp_rtp_read, ep );
	ep->trans.udp.rtcp_event =
				add_fd_event( rtcpfd, 0, 0, udp_rtcp_read, ep );
	init_rtp_endpoint( ep );

	*our_port = port;

	return ep;
}

struct rtp_endpoint *new_rtp_interleaved_endpoint( struct rtp_stream *rtp,
		void *conn, int chan )
{
	struct rtp_endpoint *ep;

	ep = (struct rtp_endpoint *)malloc( sizeof( struct rtp_endpoint ) );
	ep->rtp_stream = rtp;
	ep->trans_type = RTP_TRANS_INTER;
	ep->trans.inter.conn = conn;
	ep->trans.inter.chan = chan;
	init_rtp_endpoint( ep );

	return ep;
}

void send_rtp_packet( struct rtp_stream *rtp,
				unsigned char *d, int len, int marker )
{
	struct rtp_endpoint *ep, *next;
	unsigned char buf[1500];

	//spook_log( SL_DEBUG, "RTP: payload length %d, seq %u, time %u, marker %d",
	//		len, seqnum, timestamp, marker );

	memcpy( buf + 12, d, len );

	for( ep = rtp->ep_list; ep; ep = ep->next )
	{
		next = ep->next;

		if( ! ep->running ) continue;

		++ep->seqnum;
		ep->seqnum &= 0xFFFF;

		buf[0] = 2 << 6; /* version */
		buf[1] = ep->payload;
		if( marker ) buf[1] |= 0x80;
		PUT_16( buf + 2, ep->seqnum );
		PUT_32( buf + 4, ep->timestamp );
		PUT_32( buf + 8, ep->ssrc );

		switch( ep->trans_type )
		{
		case RTP_TRANS_UDP:
			if( send( ep->trans.udp.rtp_fd, buf, len + 12, 0 ) < 0 )
			{
				spook_log( SL_VERBOSE, "error sending UDP RTP frame: %s",
						strerror( errno ) );
				del_rtp_endpoint( ep );
				continue;
			}
			break;
		case RTP_TRANS_INTER:
			interleave_send( ep->trans.inter.conn,
					ep->trans.inter.chan, buf, len + 12 );
			break;
		}

		ep->octet_count += len;
		++ep->packet_count;

		if( ep->packet_count > 0 && ( ep->packet_count % 40 ) == 0 )
			rtcp_send( NULL, ep );
	}
}
