/*
 * Copyright (C) 2004 Nathan Lutchansky <lutchann@litech.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <event.h>
#include <log.h>
#include <frame.h>
#include <stream.h>
#include <rtp.h>
#include <conf_parse.h>
#include <base64_table.h>

void write_access_log( char *path, struct sockaddr *addr, int code, char *req,
		int length, char *referer, char *user_agent );
/*
 * Connection states:
 *   CONN_STATE_START: just accepted the connection, protocol is yet unknown
 *   CONN_STATE_RTSP: might get an RTSP request or an interleaved RTP packet
 *   CONN_STATE_RTSP_REQ: definitely getting an RTSP request
 *   CONN_STATE_INTERLEAVED: definitely getting an interleaved RTP packet
 *   CONN_STATE_HTTP: definitely getting an HTTP request
 */

#define CONN_STATE_START	0
#define CONN_STATE_RTSP		1
#define CONN_STATE_RTSP_REQ	2
#define CONN_STATE_INTERLEAVED	3
#define CONN_STATE_HTTP		10

struct conn {
	struct conn *next;
	struct conn *prev;
	int fd;
	int second_fd;
	struct sockaddr_in client_addr;
	struct event *read_event;
	struct event *second_read_event;
	struct event *write_event;
	int state;
	int base64_count;
	char http_tunnel_cookie[128];
	char req_buf[4096];
	int req_len;
	struct { char *name; char *value; } req_hdr[16];
	int interleave_len;
	unsigned char send_buf[65536];
	int send_buf_r;
	int send_buf_w;
	int drop_after;
	struct rtsp_session *attached_session;
};

struct rtsp_location {
	struct rtsp_location *next;
	struct rtsp_location *prev;
	char path[256];
	struct rtp_stream *rtp;
};

struct http_location {
	struct http_location *next;
	struct http_location *prev;
	char path[256];
	struct stream_destination *input;
	struct frame *frame;
};

struct listener {
	int fd;
};

struct conn *conn_list = NULL;
struct rtsp_session *sess_list = NULL;
struct rtsp_location *rtsp_loc_list = NULL;
struct http_location *http_loc_list = NULL;

static void do_read( struct event *e, void *d );

int new_rtsp_location( char *path, struct rtp_stream *rtp )
{
	struct rtsp_location *loc;

	loc = (struct rtsp_location *)malloc( sizeof( struct rtsp_location ) );
	loc->next = rtsp_loc_list;
	loc->prev = NULL;
	if( loc->next ) loc->next->prev = loc;
	rtsp_loc_list = loc;
	strcpy( loc->path, path );
	loc->rtp = rtp;
	return 0;
}

static void del_session( struct rtsp_session *sess )
{
	if( sess->next ) sess->next->prev = sess->prev;
	if( sess->prev ) sess->prev->next = sess->next;
	else sess_list = sess->next;
	free( sess );
}

static struct rtsp_session *new_session( void )
{
	struct rtsp_session *sess;
	int len;

	sess = (struct rtsp_session *)malloc( sizeof( struct rtsp_session ) );
	sess->next = sess_list;
	sess->prev = NULL;
	if( sess->next ) sess->next->prev = sess;
	sess_list = sess;

	sess->endpoint = NULL;
	for( len = 0; len < 15; ++len )
		sprintf( sess->id + len * 2, "%02X",
				(unsigned int)( random() & 0xff ) );
	sess->id[30] = 0;
	return sess;
}

static struct rtsp_session *get_session( char *id )
{
	struct rtsp_session *s;

	for( s = sess_list; s; s = s->next )
		if( ! strcmp( s->id, id ) ) break;
	return s;
}

static void close_session( struct rtsp_session *s )
{
	if( s->endpoint ) del_rtp_endpoint( s->endpoint );
	del_session( s );
}

static void drop_conn( struct conn *c )
{
	if( c->attached_session ) close_session( c->attached_session );
	remove_event( c->read_event );
	if( c->second_read_event ) remove_event( c->second_read_event );
	remove_event( c->write_event );
	if( c->fd >= 0 ) close( c->fd );
	c->fd = -1;
	if( c->second_fd >= 0 ) close( c->second_fd );
	c->second_fd = -1;
	if( c->next ) c->next->prev = c->prev;
	if( c->prev ) c->prev->next = c->next;
	else conn_list = c->next;
	free( c );
}

static void conn_write( struct event *e, void *d )
{
	struct conn *c = (struct conn *)d;
	int ret, len;

	while( c->send_buf_r != c->send_buf_w )
	{
		if( c->send_buf_w < c->send_buf_r )
			len = sizeof( c->send_buf ) - c->send_buf_r;
		else
			len = c->send_buf_w - c->send_buf_r;
		ret = write( c->fd, c->send_buf + c->send_buf_r, len );
		if( ret <= 0 )
		{
			if( ret < 0 && errno == EAGAIN ) return;
			drop_conn( c );
			return;
		}
		c->send_buf_r += ret;
		if( c->send_buf_r == sizeof( c->send_buf ) )
			c->send_buf_r = 0;
	}
	if( c->drop_after ) drop_conn( c );
	else set_event_enabled( c->write_event, 0 );
}

static inline int avail_send_buf( struct conn *c )
{
	if( c->send_buf_r > c->send_buf_w )
		return c->send_buf_r - c->send_buf_w - 1;
	else return sizeof( c->send_buf ) - c->send_buf_w + c->send_buf_r - 1;
}

static int send_data( struct conn *c, unsigned char *d, int len )
{
	if( avail_send_buf( c ) < len ) return 1;

	while( --len >= 0 )
	{
		c->send_buf[c->send_buf_w++] = *(d++);
		if( c->send_buf_w == sizeof( c->send_buf ) )
			c->send_buf_w = 0;
	}

	set_event_enabled( c->write_event, 1 );

	return 0;
}

int interleave_send( void *conn, int chan, unsigned char *buf, int len )
{
	struct conn *c = (struct conn *)conn;
	unsigned char head[4];

	if( avail_send_buf( conn ) < len + 4 ) return 1;
	head[0] = '$';
	head[1] = chan;
	PUT_16( head + 2, len );
	send_data( c, head, 4 );
	send_data( c, buf, len );

	return 0;
}

static char *get_header( struct conn *c, char *name )
{
	int i;
	static char *empty = "";

	for( i = 0; c->req_hdr[i].name; ++i )
		if( ! strcasecmp( c->req_hdr[i].name, name ) )
			return c->req_hdr[i].value;
	return empty;
}

static void log_request( struct conn *c, int code, int length )
{
	write_access_log( NULL, (struct sockaddr *)&c->client_addr,
		code, c->req_buf, length, get_header( c, "referer" ),
		get_header( c, "user-agent" ) );
}

static int handle_OPTIONS( struct conn *c )
{
	unsigned char rep[8192];

	send_data( c, rep, sprintf( rep, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nPublic: DESCRIBE, SETUP, TEARDOWN, PLAY\r\n\r\n", get_header( c, "CSeq" ) ) );
	return 0;
}

static int handle_DESCRIBE( struct conn *c )
{
	const char *fmtp;
	char rep[8192], sdp[8192];
	int sdp_len, rep_len;

	fmtp = rtsp_loc_list->rtp->get_sdp( rtsp_loc_list->rtp );
	if( ! fmtp )
	{
		log_request( c, 404, 0 );
		send_data( c, rep, sprintf( rep, "RTSP/1.0 404 Not Found\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) ) );
		return 0;
	}
	sdp_len = sprintf( sdp,
		"v=0\r\no=- 1 1 IN IP4 127.0.0.1\r\ns=Test\r\na=type:broadcast\r\na=control:rtsp://192.168.16.21:7080/test/\r\nt=0 0\r\nc=IN IP4 0.0.0.0\r\nm=video 0 RTP/AVP 96\r\na=rtpmap:96 MP4V-ES/90000\r\n%sa=control:track1\r\n\r\n", fmtp );
	rep_len = sprintf( rep,
		"RTSP/1.0 200 OK\r\nCSeq: %s\r\nContent-Type: application/sdp\r\nContent-Length: %d\r\n\r\n", get_header( c, "CSeq" ), sdp_len );
	log_request( c, 200, sdp_len );
	send_data( c, rep, rep_len );
	send_data( c, sdp, sdp_len );
	return 0;
}

static struct rtsp_session *rtsp_udp_setup( struct conn *c, char *t )
{
	char *p, *end;
	int cport, server_port;
	struct rtsp_session *s;
	unsigned char rep[8192];

	if( ! ( p = strstr( t, "client_port" ) ) || *(p + 11) != '=' )
		return NULL;
	cport = strtol( p + 12, &end, 10 );
	if( end == p + 12 ) return NULL;

	s = new_session();
	s->endpoint = new_rtp_udp_endpoint( rtsp_loc_list->rtp,
			c->client_addr.sin_addr, cport, &server_port );
	s->endpoint->session = s;
	spook_log( SL_VERBOSE, "our port is %d, client port is %d",
			server_port, cport );
	c->attached_session = s;
	log_request( c, 200, 0 );
	send_data( c, rep, sprintf( rep, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nSession: %s\r\nTransport: RTP/AVP;unicast;client_port=%d-%d;server_port=%d-%d\r\n\r\n", get_header( c, "CSeq" ), s->id, cport, cport + 1, server_port, server_port + 1 ) );

	return s;
}

static struct rtsp_session *rtsp_interleave_setup( struct conn *c, char *t )
{
	char *p, *end;
	int chan = 0;
	struct rtsp_session *s;
	unsigned char rep[8192];

	if( ( p = strstr( t, "interleaved" ) ) )
	{
		if( *(p + 11) != '=' ) return NULL;
		chan = strtol( p + 12, &end, 10 );
		if( end == p + 12 ) return NULL;
	}
	spook_log( SL_VERBOSE, "interleave channel is %d", chan );
	s = new_session();
	s->endpoint = new_rtp_interleaved_endpoint( rtsp_loc_list->rtp, c, chan );
	s->endpoint->session = s;
	c->attached_session = s;
	log_request( c, 200, 0 );
	send_data( c, rep, sprintf( rep, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nSession: %s\r\nTransport: RTP/AVP/TCP;unicast;interleaved=%d-%d\r\n\r\n", get_header( c, "CSeq" ), s->id, chan, chan + 1 ) );
	return s;
}

static int handle_SETUP( struct conn *c )
{
	char *t;
	struct rtsp_session *s;
	unsigned char rep[8192];

	t = get_header( c, "Transport" );

	spook_log( SL_VERBOSE, "setting up RTP (%s)", t );

	if( ! strncasecmp( t, "RTP/AVP/TCP", 11 ) )
	{
		if( ( s = rtsp_interleave_setup( c, t ) ) ) return 0;
	} else if( ( ! strncasecmp( t, "RTP/AVP", 7 ) && t[7] != '/' ) ||
			! strncasecmp( t, "RTP/AVP/UDP", 11 ) )
	{
		if( ( s = rtsp_udp_setup( c, t ) ) ) return 0;
	}
	log_request( c, 461, 0 );
	send_data( c, rep, sprintf( rep, "RTSP/1.0 461 Unsupported Transport\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) ) );
	return 0;
}

static int handle_PLAY( struct conn *c )
{
	struct rtsp_session *s;
	unsigned char rep[8192];

	if( ! ( s = get_session( get_header( c, "Session" ) ) ) ||
		! s->endpoint )
	{
		log_request( c, 454, 0 );
		send_data( c, rep, sprintf( rep, "RTSP/1.0 454 Session Not Found\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) ) );
		return 0;
	}
	log_request( c, 200, 0 );
	send_data( c, rep, sprintf( rep, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nSession: %s\r\n\r\n", get_header( c, "CSeq" ), s->id ) );
	if( ! s->endpoint->running )
	{
		spook_log( SL_VERBOSE, "starting streaming for session %s",
				s->id );
		s->endpoint->running = 1;
		s->endpoint->rtp_stream->new_endpoint( s->endpoint->rtp_stream,
				s->endpoint );
	}
	return 0;
}

static int handle_TEARDOWN( struct conn *c )
{
	struct rtsp_session *s;
	unsigned char rep[8192];

	if( ! ( s = get_session( get_header( c, "Session" ) ) ) )
	{
		log_request( c, 454, 0 );
		send_data( c, rep, sprintf( rep, "RTSP/1.0 454 Session Not Found\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) ) );
		return 0;
	}
	spook_log( SL_VERBOSE, "tearing down session %s", s->id );
	log_request( c, 200, 0 );
	send_data( c, rep, sprintf( rep, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nSession: %s\r\n\r\n", get_header( c, "CSeq" ), s->id ) );
	if( c->attached_session == s ) c->attached_session = NULL;
	close_session( s );
	return 0;
}

static void handle_QT_tunnel( struct conn *c, int post )
{
	unsigned char rep[8192];
	char *t;
	struct conn *cg;

	if( ! ( t = get_header( c, "x-sessioncookie" ) ) ||
		strlen( t ) + 1 > sizeof( c->http_tunnel_cookie ) )
	{
		send_data( c, rep, sprintf( rep, "HTTP/1.0 400 Missing session cookie\r\nConnection: close\r\nContent-Type: text/html\r\n\r\n<html><body><p>Broken RTSP/HTTP tunnel request</p></body></html>\n" ) );
		c->drop_after = 1;
		return;
	}
	if( ! post )
	{
		strcpy( c->http_tunnel_cookie, t );
		spook_log( SL_VERBOSE,
				"processing RTSP-HTTP tunnel GET for %s", t );
		send_data( c, rep, sprintf( rep, "HTTP/1.0 200 OK\r\nConnection: close\r\nCache-Control: no-store\r\nPragma: no-cache\r\nContent-Type: application/x-rtsp-tunnelled\r\n\r\n" ) );
		c->state = CONN_STATE_RTSP;
		c->base64_count = 0;
		return;
	}
	for( cg = conn_list; cg; cg = cg->next )
		if( ! strcmp( cg->http_tunnel_cookie, t ) ) break;
	if( ! cg )
	{
		spook_log( SL_VERBOSE,
			"unable to locate RTSP-HTTP GET request for %s", t );
		send_data( c, rep, sprintf( rep, "HTTP/1.0 400 Unknown session cookie\r\nConnection: close\r\nContent-Type: text/html\r\n\r\n<html><body><p>Broken RTSP/HTTP tunnel request</p></body></html>\n" ) );
		c->drop_after = 1;
		return;
	}
	spook_log( SL_VERBOSE, "RTSP-HTTP tunnel established for %s", t );
	if( cg->second_fd >= 0 ) close( cg->second_fd );
	if( cg->second_read_event ) remove_event( cg->second_read_event );
	cg->second_fd = c->fd;
	cg->second_read_event = add_fd_event( c->fd, 0, 0, do_read, cg );
	c->fd = -1;
	drop_conn( c ); /* doesn't close the TCP socket */
}

static int handle_GET( struct conn *c )
{
	struct http_location *loc = http_loc_list;
	unsigned char rep[8192];
	char *t;

	if( ( t = get_header( c, "Accept" ) ) &&
			! strcasecmp( t, "application/x-rtsp-tunnelled" ) )
	{
		handle_QT_tunnel( c, 0 );
		return 0;
	}

	if( ! loc || ! loc->frame || avail_send_buf( c ) < loc->frame->length + 400 )
	{
		log_request( c, 404, 0 );
		send_data( c, rep, sprintf( rep, "HTTP/1.0 404 Not Found\r\nConnection: close\r\nContent-Type: text/html\r\n\r\n<html><body><p>Image not available</p></body></html>\n" ) );
		c->drop_after = 1;
		return 0;
	}
	log_request( c, 200, loc->frame->length );
	send_data( c, rep, sprintf( rep, "HTTP/1.0 200 OK\r\nConnection: close\r\nContent-Type: image/jpeg\r\nContent-Length: %d\r\n\r\n", loc->frame->length ) );
	send_data( c, loc->frame->d, loc->frame->length );
	c->drop_after = 1;
	return 0;
}

static int handle_POST( struct conn *c )
{
	unsigned char rep[8192];
	char *t;

	if( ( t = get_header( c, "Content-Type" ) ) &&
			! strcasecmp( t, "application/x-rtsp-tunnelled" ) )
	{
		handle_QT_tunnel( c, 1 );
		return -1;
	}

	log_request( c, 501, 0 );
	send_data( c, rep, sprintf( rep, "HTTP/1.0 501 Not Implemented\r\nConnection: close\r\nContent-Type: text/html\r\n\r\n<html><body><p>POST requests are unsupported here</p></body></html>\n" ) );
	c->drop_after = 1;
	return 0;
}

static int handle_unknown( struct conn *c )
{
	unsigned char rep[8192];

	log_request( c, 501, 0 );
	send_data( c, rep, sprintf( rep, "RTSP/1.0 501 Not Implemented\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) ) );
	return 0;
}

static int handle_request( struct conn *c )
{
	char *t;
	int i = 0;

	if( c->req_buf[0] == 0 ) return -1;

	/* Terminate the first line of the request */
	for( t = c->req_buf; *t != '\r' && *t != '\n'; ++t )
		if( *t == 0 ) return -1;
	*(t++) = 0;
	if( *t == '\n' ) *(t++) = 0;
	/* Find the name and content of each header line */
	for( i = 0; i < 15; ++i )
	{
		/* We should be sitting on the header name */
		c->req_hdr[i].name = t;
		/* Terminate where the colon is */
		for( ; *t != ':'; ++t )
			if( *t == 0 || *t == '\r' || *t == '\n' ) return -1;
		*(t++) = 0;
		/* Skip any spaces after the colon */
		for( ; *t == ' '; ++t )
			if( *t == 0 ) return -1;
		/* Should be the start of the header content */
		c->req_hdr[i].value = t;
		/* Terminate at the end of the line */
		for( ; *t != '\r' && *t != '\n'; ++t )
			if( *t == 0 ) return -1;
		*(t++) = 0;
		if( *t == '\n' ) *(t++) = 0;
		/* Are we at the end of the header? */
		if( *t == 0 || *t == '\r' || *t == '\n' ) break;
	}
	if( i == 15 ) return -1; /* Too many headers */
	c->req_hdr[i + 1].name = NULL;

#if 0
	for( i = 0; i < 16 && c->req_hdr[i].name; ++i )
	{
		spook_log( SL_DEBUG, "Header name = %s\n", c->req_hdr[i].name );
		spook_log( SL_DEBUG, "Header value = %s\n", c->req_hdr[i].value );
	}
#endif

	spook_log( SL_VERBOSE, "client request: %s", c->req_buf );

	if( ! strncasecmp( c->req_buf, "OPTIONS ", 8 ) )
		return handle_OPTIONS( c );
	else if( ! strncasecmp( c->req_buf, "DESCRIBE ", 9 ) )
		return handle_DESCRIBE( c );
	else if( ! strncasecmp( c->req_buf, "SETUP ", 6 ) )
		return handle_SETUP( c );
	else if( ! strncasecmp( c->req_buf, "PLAY ", 5 ) )
		return handle_PLAY( c );
	else if( ! strncasecmp( c->req_buf, "TEARDOWN ", 9 ) )
		return handle_TEARDOWN( c );
	else if( ! strncasecmp( c->req_buf, "GET ", 4 ) )
		return handle_GET( c );
	else if( ! strncasecmp( c->req_buf, "POST ", 5 ) )
		return handle_POST( c );
	else return handle_unknown( c );
}

static int parse_client_data( struct conn *c )
{
	char *t;

	switch( c->state )
	{
	case CONN_STATE_START:
		if( strchr( c->req_buf, '\n' ) )
		{
			if( ! strncasecmp( c->req_buf, "GET", 3 ) )
				c->state = CONN_STATE_HTTP;
			else
				c->state = CONN_STATE_RTSP;
			return 1;
		}
		break;
	case CONN_STATE_RTSP:
		if( c->req_len < 4 ) return 0;
		if( c->req_buf[0] == '$' )
		{
			c->state = CONN_STATE_INTERLEAVED;
			c->interleave_len = GET_16( c->req_buf + 2 );
			if( c->interleave_len + 4 > sizeof( c->req_buf ) )
				return -1;
		} else c->state = CONN_STATE_RTSP_REQ;
		return 1;
	case CONN_STATE_HTTP:
	case CONN_STATE_RTSP_REQ:
		if( ! ( t = strstr( c->req_buf, "\r\n\r\n" ) ) ) return 0;
		t[2] = 0;
		if( handle_request( c ) < 0 ) return -1;
		/* should really just delete the request from the
		 * buffer, not the entire buffer */
		c->req_len = 0;
		if( c->state == CONN_STATE_RTSP_REQ )
			c->state = CONN_STATE_RTSP;
		break;
	case CONN_STATE_INTERLEAVED:
		if( c->req_len < c->interleave_len + 4 ) return 0;
		/* should really just delete the packet from the
		 * buffer, not the entire buffer */
		c->state = CONN_STATE_RTSP;
		c->req_len = 0;
		break;
	}
	return 0;
}

static int unbase64( unsigned char *d, int len, int *remain )
{
	int src = 0, dest = 0, i;
	unsigned char enc[4];

	for(;;)
	{
		for( i = 0; i < 4; ++src )
		{
			if( src >= len )
			{
				if( i > 0 ) memcpy( d + dest, enc, i );
				*remain = i;
				return dest + i;
			}
			if( base64[d[src]] < 0 ) continue;
			enc[i++] = d[src];
		}
		d[dest] = ( ( base64[enc[0]] << 2 ) & 0xFC ) |
				( base64[enc[1]] >> 4 );
		d[dest + 1] = ( ( base64[enc[1]] << 4 ) & 0xF0 ) |
				( base64[enc[2]] >> 2 );
		d[dest + 2] = ( ( base64[enc[2]] << 6 ) & 0xC0 ) |
				base64[enc[3]];
		if( enc[3] == '=' )
		{
			if( enc[2] == '=' ) dest += 1;
			else dest += 2;
		} else dest += 3;
	}
}

static void do_read( struct event *e, void *d )
{
	struct conn *c = (struct conn *)d;
	int ret, second;

	second = e == c->second_read_event;
	for(;;)
	{
		ret = read( second ? c->second_fd : c->fd,
				c->req_buf + c->req_len,
				sizeof( c->req_buf ) - c->req_len - 1 );
		if( ret <= 0 )
		{
			if( ret < 0 )
			{
				if( errno == EAGAIN ) return;
				spook_log( SL_VERBOSE, "closing TCP connection to client due to read error: %s",
						strerror( errno ) );
			} else spook_log( SL_VERBOSE, "client closed TCP connection" );
			if( second )
			{
				remove_event( c->second_read_event );
				c->second_read_event = NULL;
				close( c->second_fd );
				c->second_fd = -1;
			} else drop_conn( c );
			return;
		}

		if( c->base64_count >= 0 )
		{
			c->req_len -= c->base64_count;
			c->req_len +=
				unbase64( c->req_buf + c->req_len,
						ret + c->base64_count,
						&c->base64_count );
		} else c->req_len += ret;

		if( c->req_len == sizeof( c->req_buf ) - 1 )
		{
			spook_log( SL_VERBOSE, "malformed request from client; exceeded maximum size" );
			drop_conn( c );
			return;
		}

		if( c->base64_count > 0 ) continue;

		c->req_buf[c->req_len] = 0;
		while( ( ret = parse_client_data( c ) ) > 0 );
		if( ret < 0 ) return;
	}
}

static void do_accept( struct event *e, void *d )
{
	struct listener *listener = (struct listener *)d;
	int fd;
	int addrlen;
	struct sockaddr_in addr;
	struct conn *c;

	addrlen = sizeof( addr );
	if( ( fd = accept( listener->fd, (struct sockaddr *)&addr, &addrlen ) ) < 0 )
	{
		spook_log( SL_WARN, "error accepting TCP connection: %s",
				strerror( errno ) );
		return;
	}
	spook_log( SL_VERBOSE, "accepted connection from %s:%d",
			inet_ntoa( addr.sin_addr ), ntohs( addr.sin_port ) );

	fcntl( fd, F_SETFL, O_NONBLOCK );

	c = (struct conn *)malloc( sizeof( struct conn ) );
	c->next = conn_list;
	if( c->next ) c->next->prev = c;
	c->prev = NULL;
	c->fd = fd;
	c->second_fd = -1;
	c->client_addr = addr;
	c->state = CONN_STATE_START;
	c->base64_count = -1;
	c->req_len = 0;
	c->read_event = add_fd_event( fd, 0, 0, do_read, c );
	c->second_read_event = NULL;
	c->write_event = add_fd_event( fd, 1, 0, conn_write, c );
	set_event_enabled( c->write_event, 0 );
	c->send_buf_r = c->send_buf_w = 0;
	c->drop_after = 0;
	c->attached_session = NULL;
	conn_list = c;
}

int rtsp_listen( int port )
{
	struct sockaddr_in addr;
	struct listener *listener;
	int opt, fd;

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = 0;
	addr.sin_port = htons( port );

	if( ( fd = socket( PF_INET, SOCK_STREAM, 0 ) ) < 0 )
	{
		spook_log( SL_ERR, "error creating listen socket: %s",
				strerror( errno ) );
		return -1;
	}
	opt = 1;
	if( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof( opt ) ) < 0 )
		spook_log( SL_WARN, "ignoring error on setsockopt: %s",
				strerror( errno ) );
	if( bind( fd, (struct sockaddr *)&addr, sizeof( addr ) ) < 0 )
	{
		spook_log( SL_ERR, "unable to bind to tcp socket: %s",
				strerror( errno ) );
		close( fd );
		return -1;
	}
	if( listen( fd, 5 ) < 0 )
	{
		spook_log( SL_ERR,
			"error when attempting to listen on tcp socket: %s",
			strerror( errno ) );
		close( fd );
		return -1;
	}

	listener = (struct listener *)malloc( sizeof( struct listener ) );
	listener->fd = fd;

	add_fd_event( fd, 0, 0, do_accept, listener );

	spook_log( SL_INFO, "listening on tcp port %d", port );

	return 0;
}

/********************* GLOBAL CONFIGURATION DIRECTIVES ********************/

int config_port( int num_tokens, struct token *tokens, void *d )
{
	int port;

	port = tokens[1].v.num;

	if( port <= 0 || port > 65535 )
	{
		spook_log( SL_ERR, "invalid listen port %d", port );
		return -1;
	}

	return rtsp_listen( port );
}

/************************ CONFIGURATION DIRECTIVES ************************/

static void jpeg_next_frame( struct frame *f, void *d )
{
	struct http_location *loc = (struct http_location *)d;

	if( loc->frame ) unref_frame( loc->frame );
	loc->frame = f;
}

static void *start_block(void)
{
	struct http_location *loc;

	loc = (struct http_location *)malloc( sizeof( struct http_location ) );
	loc->next = http_loc_list;
	loc->prev = NULL;
	if( loc->next ) loc->next->prev = loc;
	http_loc_list = loc;
	loc->input = NULL;
	loc->path[0] = 0;
	loc->frame = NULL;

	return loc;
}

static int end_block( void *d )
{
	struct http_location *loc = (struct http_location *)d;

	if( ! loc->input )
	{
		spook_log( SL_ERR, "http output: missing input stream name" );
		return -1;
	}
	if( ! *loc->path )
	{
		spook_log( SL_ERR, "http output: missing URL path" );
		return -1;
	}
	set_waiting( loc->input, 1 );

	return 0;
}

static int set_input( int num_tokens, struct token *tokens, void *d )
{
	struct http_location *loc = (struct http_location *)d;
	int format = FORMAT_JPEG;

	if( ! ( loc->input = connect_to_stream( tokens[1].v.str,
					jpeg_next_frame, loc, &format, 1 ) ) )
	{
		spook_log( SL_ERR,
			"http output: unable to connect to stream \"%s\"",
			tokens[1].v.str );
		return -1;
	}
	return 0;
}

static int set_path( int num_tokens, struct token *tokens, void *d )
{
	struct http_location *loc = (struct http_location *)d;

	strcpy( loc->path, tokens[1].v.str );
	return 0;
}

static struct statement config_statements[] = {
	/* directive name, process function, min args, max args, arg types */
	{ "input", set_input, 1, 1, { TOKEN_STR } },
	{ "path", set_path, 1, 1, { TOKEN_STR } },

	/* empty terminator -- do not remove */
	{ NULL, NULL, 0, 0, {} }
};

int http_init(void)
{
	register_config_context( "output", "http", start_block, end_block,
					config_statements );
	return 0;
}
