/*
 * Copyright (C) 2004 Nathan Lutchansky <lutchann@litech.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <event.h>
#include <log.h>
#include <frame.h>
#include <stream.h>
#include <conversions.h>

static struct stream *stream_list = NULL;

static int is_valid_format( struct stream_destination *dest, int format )
{
	int i;

	for( i = 0; i < dest->format_option_count; ++i )
		if( dest->format_options[i] == format ) return 1;
	return 0;
}

static int choose_dest_format( struct stream_destination *dest, int format )
{
	if( is_valid_format( dest, format ) )
		dest->format = format;
	else switch( format )
	{
	case FORMAT_RAW_UYVY:
		if( is_valid_format( dest, FORMAT_RAW_RGB24 ) )
			dest->format = FORMAT_RAW_RGB24;
		else if( is_valid_format( dest, FORMAT_RAW_BGR24 ) )
			dest->format = FORMAT_RAW_BGR24;
		break;
	case FORMAT_RAW_RGB24:
		if( is_valid_format( dest, FORMAT_RAW_BGR24 ) )
			dest->format = FORMAT_RAW_BGR24;
		break;
	case FORMAT_RAW_BGR24:
		if( is_valid_format( dest, FORMAT_RAW_RGB24 ) )
			dest->format = FORMAT_RAW_RGB24;
		break;
	}
	if( dest->format > 0 ) return 1;
	spook_log( SL_WARN,
		"can't convert format %d to anything usable", format );
	return 0;
}

void deliver_frame_to_stream( struct frame *f, void *d )
{
	struct stream *s = (struct stream *)d;
	struct stream_destination *dest;
	struct frame *uyvy = NULL, *rgb = NULL, *bgr = NULL;
	int i;

	switch( f->format )
	{
	case FORMAT_RAW_UYVY:
		uyvy = f;
		break;
	case FORMAT_RAW_RGB24:
		rgb = f;
		break;
	case FORMAT_RAW_BGR24:
		bgr = f;
		break;
	}

	for( dest = s->dest_list; dest; dest = dest->next )
	{
		if( ! dest->waiting ) continue;
		if( dest->format < 0 &&
				! choose_dest_format( dest, f->format ) )
			continue;
		switch( dest->format )
		{
		case FORMAT_RAW_UYVY:
			if( ! uyvy && rgb )
			{
				/* convert */
				spook_log( SL_WARN,
					"not yet converting from RGB to UYUV" );
			}
			if( uyvy )
			{
				ref_frame( uyvy );
				dest->process_frame( uyvy, dest->d );
			} else spook_log( SL_WARN, "can't convert format %d to %d",
						f->format, dest->format );
			break;
		case FORMAT_RAW_RGB24:
			if( ! rgb && bgr )
			{
				rgb = new_frame();
				rgb->format = FORMAT_RAW_RGB24;
				rgb->width = bgr->width;
				rgb->height = bgr->height;
				rgb->length = bgr->height * bgr->width * 3;
				rgb->key = bgr->key;
				for( i = 0; i < rgb->length; i += 3 )
				{
					rgb->d[i] = bgr->d[i+2];
					rgb->d[i+1] = bgr->d[i+1];
					rgb->d[i+2] = bgr->d[i];
				}
			}
			if( ! rgb && uyvy )
			{
				rgb = new_frame();
				rgb->format = FORMAT_RAW_RGB24;
				rgb->width = uyvy->width;
				rgb->height = uyvy->height;
				rgb->length = uyvy->height * uyvy->width * 3;
				rgb->key = uyvy->key;
				uyvy2rgb( uyvy->d, rgb->d, uyvy->length / 2 );
			}
			if( rgb )
			{
				ref_frame( rgb );
				dest->process_frame( rgb, dest->d );
			} else spook_log( SL_WARN, "can't convert format %d to %d",
						f->format, dest->format );
			break;
		case FORMAT_RAW_BGR24:
			if( ! bgr && rgb )
			{
				bgr = new_frame();
				bgr->format = FORMAT_RAW_BGR24;
				bgr->width = rgb->width;
				bgr->height = rgb->height;
				bgr->length = rgb->height * rgb->width * 3;
				bgr->key = rgb->key;
				for( i = 0; i < bgr->length; i += 3 )
				{
					bgr->d[i] = rgb->d[i+2];
					bgr->d[i+1] = rgb->d[i+1];
					bgr->d[i+2] = rgb->d[i];
				}
			}
			if( bgr )
			{
				ref_frame( bgr );
				dest->process_frame( bgr, dest->d );
			} else spook_log( SL_WARN, "can't convert format %d to %d",
						f->format, dest->format );
			break;
		default:
			if( dest->format == 0 || f->format == dest->format )
			{
				ref_frame( f );
				dest->process_frame( f, dest->d );
			} else spook_log( SL_WARN, "can't convert format %d to %d",
						f->format, dest->format );
			break;
		}
	}

	if( rgb && rgb != f ) unref_frame( rgb );
	if( bgr && bgr != f ) unref_frame( bgr );
	if( uyvy && uyvy != f ) unref_frame( uyvy );
	unref_frame( f );
}

struct stream_destination *connect_to_stream( char *name,
		frame_deliver_func process_frame, void *d,
		int *formats, int format_count )
{
	struct stream *s;
	struct stream_destination *dest;

	for( s = stream_list; s; s = s->next )
		if( ! strcmp( s->name, name ) ) break;

	if( ! s ) return NULL;

	dest = (struct stream_destination *)
			malloc( sizeof( struct stream_destination ) );
	dest->next = s->dest_list;
	dest->prev = NULL;
	if( dest->next ) dest->next->prev = dest;
	s->dest_list = dest;

	dest->format_option_count = format_count;
	if( format_count > 0 )
	{
		dest->format_options = (int *)
			malloc( format_count * sizeof( int ) );
		memcpy( dest->format_options, formats,
				format_count * sizeof( int ) );
		dest->format = -1;
	} else
	{
		dest->format_options = NULL;
		dest->format = 0;
	}
	dest->stream = s;
	dest->waiting = 0;
	dest->process_frame = process_frame;
	dest->d = d;

	return dest;
}

void del_stream( struct stream *s )
{
	if( s->next ) s->next->prev = s->prev;
	if( s->prev ) s->prev->next = s->next;
	else stream_list = s->next;
	free( s );
}

struct stream *new_stream( char *name, void *d )
{
	struct stream *s;

	s = (struct stream *)malloc( sizeof( struct stream ) );
	s->next = stream_list;
	s->prev = NULL;
	if( s->next ) s->next->prev = s;
	stream_list = s;
	strcpy( s->name, name );
	s->dest_list = NULL;
	s->set_running = NULL;
	s->private = d;
	return s;
}

void set_waiting( struct stream_destination *dest, int waiting )
{
	struct stream *s = dest->stream;

	/* We call set_running every time a destination starts listening,
	 * or when the last destination stops listening.  It is good to know
	 * when new listeners come so maybe the source can send a keyframe. */

	if( dest->waiting )
	{
		if( waiting ) return; /* no change in status */
		dest->waiting = 0;

		/* see if anybody else is listening */
		for( dest = s->dest_list; dest; dest = dest->next )
			waiting |= dest->waiting;
		if( waiting ) return; /* others are still listening */
	} else
	{
		if( ! waiting ) return; /* no change in status */
		dest->waiting = 1;
	}

	s->set_running( s, waiting );
}
